#############################################################################
##  
#W  lib/Declare.g                      FUtil                    Frank Lbeck 
##  
#Y  Copyright (C)  2004  
##  
##  Here  we   define  simplified  commands  for   declarations  and  method
##  installations.
##  
FUtil.StoreRevision( "lib/Declare.g",
   "$Id: Declare.g 8 2004-05-27 00:08:19Z luebeck $" );

# to use this file in several places:
if not IsBound(NewMethod) then
  
  
##  
##  This file contains some shortcuts for 'DeclareXXX' and method
##  installation commands.
##  

##  For new objects we want to introduce, we will not use families.
##  Unfortunately the family argument in 'NewType' is not optional.
##  Therefore:
##  
BindGlobal("DefaultFamily", NewFamily("DefaultFamily"));

BindGlobal("MakeType", function(arg)
  return CallFuncList(NewType, Concatenation([DefaultFamily], arg));
end);

##  We don't want to introduce restrictions on the arguments during the
##  declaration, which can be annoying for other users who want to install
##  methods for it. 

##  for each string in the argument an attribute, resp. operation is declared
BindGlobal("MakeAttribute", function(arg)
  local str;
  for str  in arg do
    DeclareAttribute(str, IsObject);
  od;
end);
BindGlobal("MakeOperation", function(arg)
  local str;
  for str in arg do
    DeclareOperation(str, []);
  od;
end);
BindGlobal("MakeGlobalVariable", function(arg)
  local str;
  for str in arg do
    DeclareGlobalVariable(str);
  od;
end);
BindGlobal("MakeGlobalFunction", function(arg)
  local str;
  for str in arg do
    DeclareGlobalFunction(str);
  od;
end);

##  contrary to 'DeclareProperty' this installs impl as implied filters
BindGlobal("MakeProperty", function(arg)
  local   str, impl, rk, pr;
  str := arg[1];
  impl := arg[2];
  if Length(arg)>2 then
    rk := arg[3];
  else
    rk := 0;
  fi;
  DeclareProperty(str, impl, rk);
  pr := ValueGlobal(str);
  InstallTrueMethod(impl, pr);
  InstallOtherMethod(Setter(pr), "true if missing", true, [IsObject], 0,
          function(obj)
    Setter(pr)(obj, true);
  end);
end);


##  
##  We simplify the method installation. We don't want to use the families,
##  hope that we are not forced to install a method with a rank offset 
##  <> 0, we generate the comment argument automatically; the filter can
##  contain "or" conjunctions which leads to several method installations.
##  
##  arguments: name, filter, fun[, offset] 
##  (the fourth argument is hopefully at most temporarily used to 
##  go around problems from library)
BindGlobal("NewMethod", function(arg)
  local   c3,  name,  filter,  fun,  offset,  i,  cname,  snames,  
          sname,  str;
  
  # since `Cartesian3' may not yet be loaded
  c3 := function(arg) return Cartesian2( arg, Length( arg ), [  ], 1); end;
  
  name := arg[1];
  filter := arg[2];
  fun := arg[3];
  if LEN_LIST(arg)>3 then
    offset := arg[4];
  else
    offset := 0;
  fi;
  
  if IsString(filter) or not IsList(filter) then
    filter := [filter];
  fi;
  ##  filters not given as string are converted to string (which does
  ##  not work for `IsObject')
  for i in [1..Length(filter)] do
    if not IsString(filter[i]) then
      if filter[i] = IsObject then
        filter[i] := "IsObject";
      else
        filter[i] := JoinStringsWithSeparator(NamesFilter(filter[i]),
                             " and ");
      fi;
    fi;
  od;
  cname := JoinStringsWithSeparator(filter);
  
  # now we split filters for each argument along the `or'
  filter := List(filter, a-> SplitString(
                    ReplacedString(a, " or ", "%"), "%"));
  
  snames := List(CallFuncList(c3, filter), a->
                 JoinStringsWithSeparator(a));
  
  # thats it, we can now install the methods:
  for sname in snames do
    # evaluating the strings for the filters
    filter := EvalString(Concatenation("[", sname, "]"));
  
    # we always call `InstallOtherMethod' since we don't want to be bothered
    # by possibly to strong conditions on the arguments in the declaration
    # of the operation
    InstallOtherMethod(name, cname, true, filter, offset, fun);
  od;
end);


# introducing a display of component objects
NewMethod(Display, ["IsRecord or IsComponentObjectRep", "IsInt"], 
function(r, ind)
  local cc, indstr;
  indstr := ListWithIdenticalEntries(ind, ' ');
  for cc in NamesOfComponents(r) do
    Print(indstr, cc, ": ");
    Display(r!.(cc), ind+2);
  od;
end);
NewMethod(Display, ["IsObject", "IsInt"], function(a, ind) Display(a); end);
NewMethod(Display, ["IsRecord or IsComponentObjectRep"], 
function(a) Display(a, 0); end);


fi;
